@@ -7,16 +7,29 @@ eval "use WebminCore;";
 &foreign_require("software", "software-lib.pl");
 &foreign_require("cron", "cron-lib.pl");
 &foreign_require("webmin", "webmin-lib.pl");
-use Data::Dumper;
 
-$available_cache_file = "$module_config_directory/available.cache";
-$current_cache_file = "$module_config_directory/current.cache";
-$updates_cache_file = "$module_config_directory/updates.cache";
+$available_cache_file = &cache_file_path("available.cache");
+$current_cache_file = &cache_file_path("current.cache");
+$updates_cache_file = &cache_file_path("updates.cache");
 $cron_cmd = "$module_config_directory/update.pl";
 
-$yum_cache_file = "$module_config_directory/yumcache";
-$apt_cache_file = "$module_config_directory/aptcache";
-$yum_changelog_cache_dir = "$module_config_directory/yumchangelog";
+$yum_cache_file = &cache_file_path("yumcache");
+$apt_cache_file = &cache_file_path("aptcache");
+$yum_changelog_cache_dir = &cache_file_path("yumchangelog");
+
+$update_progress_dir = "$module_var_directory/progress";
+
+# cache_file_path(name)
+# Returns a path in the /var directory unless the file already exists under
+# /etc/webmin
+sub cache_file_path
+{
+my ($name) = @_;
+if (-e "$module_config_directory/$name") {
+	return "$module_config_directory/$name";
+	}
+return "$module_var_directory/$name";
+}
 
 # get_software_packages()
 # Fills in software::packages with list of installed packages (if missing),
@@ -25,12 +38,15 @@ sub get_software_packages
 {
 if (!$get_software_packages_cache) {
         %software::packages = ( );
+	if (!defined(&software::list_packages)) {
+		return 0;
+		}
         $get_software_packages_cache = &software::list_packages();
         }
 return $get_software_packages_cache;
 }
 
-# list_current(nocache)
+# list_current([nocache])
 # Returns a list of packages and versions for installed software. Keys are :
 #  name - The my package name (ie. CSWapache2)
 #  update - Name used to refer to it by the updates system (ie. apache2)
@@ -139,10 +155,13 @@ return 0;
 sub write_cache_file
 {
 my ($file, $arr) = @_;
-&open_tempfile(FILE, ">$file");
-&print_tempfile(FILE, Dumper($arr));
-&close_tempfile(FILE);
-$read_cache_file_cache{$file} = $arr;
+eval "use Data::Dumper";
+if (!$@) {
+	&open_tempfile(FILE, ">$file");
+	&print_tempfile(FILE, Dumper($arr));
+	&close_tempfile(FILE);
+	$read_cache_file_cache{$file} = $arr;
+	}
 }
 
 # read_cache_file(file)
@@ -161,7 +180,7 @@ if (-r $file) {
 return ( );
 }
 
-# compare_versions(&pkg1, &pk2)
+# compare_versions(&pkg1, &pkg2)
 # Returns -1 if the version of pkg1 is older than pkg2, 1 if newer, 0 if same.
 sub compare_versions
 {
@@ -173,7 +192,7 @@ if ($pkg1->{'system'} eq 'webmin' && $pkg2->{'system'} eq 'webmin') {
 my $ec = $pkg1->{'epoch'} <=> $pkg2->{'epoch'};
 if ($ec && ($pkg1->{'epoch'} eq '' || $pkg2->{'epoch'} eq '') &&
     $pkg1->{'system'} eq 'apt') {
-	# On some Debian systems, we don't have a my epoch
+	# On some Debian systems, we don't have any epoch
 	$ec = undef;
 	}
 return $ec ||
@@ -275,12 +294,13 @@ if (!scalar(@updates_available_cache)) {
 return @updates_available_cache;
 }
 
-# package_install(package-name, [system])
+# package_install(package-name, [system], [new-install], [flags])
 # Install some package, either from an update system or from Webmin. Returns
 # a list of updated package names.
 sub package_install
 {
-my ($name, $system) = @_;
+my ($name, $system, $install, $flags) = @_;
+$system ||= $software::update_system;
 my @rv;
 my $pkg;
 
@@ -296,6 +316,11 @@ if (!$pkg) {
 		      sort { &compare_versions($b, $a) }
 			   &list_available(0);
 	}
+if (!$pkg && $install) {
+	# Assume that it will exist
+	$pkg = { 'system' => $system || $software::update_system,
+		 'name' => $name };
+	}
 if (!$pkg) {
 	print &text('update_efindpkg', $name),"<p>\n";
 	return ( );
@@ -305,7 +330,19 @@ if (defined(&software::update_system_install)) {
 	&clean_environment();
 	if ($software::update_system eq $pkg->{'system'}) {
 		# Can use the default system
-		@rv = &software::update_system_install($name, undef, 1);
+		if ($name eq "apache2" &&
+		    $pkg->{'system'} eq 'apt') {
+			# If updating the apache2 package on an apt system
+			# and apache2-mpm-prefork is installed, also update it
+			# so that ubuntu doesn't pull in the apache2-mpm-worker
+			# instead, which breaks PHP :-(
+			local @pinfo = &software::package_info(
+					"apache2-mpm-prefork");
+			if (@pinfo) {
+				$name .= " apache2-mpm-prefork";
+				}
+			}
+		@rv = &software::update_system_install($name, undef, 1, $flags);
 		}
 	else {
 		# Another update system exists!! Use it..
@@ -315,7 +352,7 @@ if (defined(&software::update_system_install)) {
 		if (!$done_rhn_text++) {
 			%text = ( %text, %software::text );
 			}
-		@rv = &update_system_install($name, undef, 1);
+		@rv = &update_system_install($name, undef, 1, $flags);
 		}
 	&reset_environment();
 	}
@@ -327,12 +364,13 @@ unlink($current_cache_file);
 return @rv;
 }
 
-# package_install_multiple(&package-names, system)
+# package_install_multiple(&package-names, system, [new-install], [flags])
 # Install multiple packages, either from an update system or from Webmin.
 # Returns a list of updated package names.
 sub package_install_multiple
 {
-my ($names, $system) = @_;
+my ($names, $system, $install, $flags) = @_;
+$system ||= $software::update_system;
 my @rv;
 my $pkg;
 
@@ -342,7 +380,7 @@ if (defined(&software::update_system_install)) {
 	if ($software::update_system eq $system) {
 		# Can use the default system
 		@rv = &software::update_system_install(
-			join(" ", @$names), undef, 1);
+			join(" ", @$names), undef, 1, $flags);
 		}
 	else {
 		# Another update system exists!! Use it..
@@ -352,7 +390,7 @@ if (defined(&software::update_system_install)) {
 		if (!$done_rhn_text++) {
 			%text = ( %text, %software::text );
 			}
-		@rv = &update_system_install(join(" ", @$names), undef, 1);
+		@rv = &update_system_install(join(" ", @$names), undef, 1, $flags);
 		}
 	&reset_environment();
 	}
@@ -380,15 +418,16 @@ if (defined(&software::update_system_operations)) {
 return ( );
 }
 
-# list_possible_updates([nocache])
+# list_possible_updates([nocache], [nocache-no-data])
 # Returns a list of updates that are available. Each element in the array
 # is a hash ref containing a name, version, description and severity flag.
 # Intended for calling from themes. Nocache 0=cache everything, 1=flush all
-# caches, 2=flush only current
+# caches, 2=flush only current. Nocache-no-data prohibits collecting data
 sub list_possible_updates
 {
-my ($nocache) = @_;
+my ($nocache, $nocache_no_data) = @_;
 my @rv;
+return @rv if ($nocache_no_data);
 my @current = &list_current($nocache);
 if (&supports_updates_available()) {
 	# Software module supplies a function that can list just packages
@@ -400,7 +439,8 @@ if (&supports_updates_available()) {
 	foreach my $a (&updates_available($nocache == 1)) {
 		my $c = $currentmap{$a->{'name'},$a->{'system'}};
 		next if (!$c);
-		next if ($a->{'version'} eq $c->{'version'});
+		next if ($a->{'version'} eq $c->{'version'} &&
+			 $a->{'epoch'} eq $c->{'epoch'});
 		push(@rv, { 'name' => $a->{'name'},
 			    'update' => $a->{'update'},
 			    'system' => $a->{'system'},
@@ -442,6 +482,7 @@ else {
 			}
 		}
 	}
+@rv = &filter_duplicates(\@rv);
 return @rv;
 }
 
@@ -517,11 +558,20 @@ return 1;
 sub clear_repository_cache
 {
 if ($software::update_system eq "yum") {
-	&execute_command("yum clean all");
+	&execute_command("$software::yum_command clean all");
 	}
 elsif ($software::update_system eq "apt") {
 	&execute_command("apt-get update");
 	}
+elsif ($software::update_system eq "ports") {
+	&foreign_require("proc");
+	foreach my $cmd ("portsnap fetch",
+			 "portsnap update || portsnap extract") {
+		my ($fh, $pid) = &proc::pty_process_exec($cmd);
+		while(<$fh>) { }
+		close($fh);
+		}
+	}
 }
 
 # split_epoch(version)
@@ -543,8 +593,8 @@ my ($pkg) = @_;
 if ($pkg->{'system'} eq 'yum') {
 	# See if yum supports changelog
 	if (!defined($supports_yum_changelog)) {
-		my $out = &backquote_command("yum -h 2>&1 </dev/null");
-		$supports_yum_changelog = $out =~ /changelog/ ? 1 : 0;
+		my $out = &backquote_command("$software::yum_command -h 2>&1 </dev/null");
+		$supports_yum_changelog = $out =~ /changelog|updateinfo/ ? 1 : 0;
 		}
 	return undef if (!$supports_yum_changelog);
 
@@ -552,25 +602,47 @@ if ($pkg->{'system'} eq 'yum') {
 	my $cfile = $yum_changelog_cache_dir."/".
 		       $pkg->{'name'}."-".$pkg->{'version'};
 	my $cl = &read_file_contents($cfile);
-	if (!$cl) {
-		# Run it for this package and version
+	if (!$cl && $software::yum_command =~ /yum/) {
+		# Run yum changelog for this package and version
 		my $started = 0;
-		&open_execute_command(YUMCL, "yum changelog all ".
-					     quotemeta($pkg->{'name'}), 1, 1);
-		while(<YUMCL>) {
+		&open_execute_command(YUMCL,
+			"$software::yum_command changelog all ".
+		        quotemeta($pkg->{'name'}), 1, 1);
+                while(<YUMCL>) {
+                        s/\r|\n//g;
+                        if (/^\Q$pkg->{'name'}-$pkg->{'version'}\E/) {
+                                $started = 1;
+                                }
+                        elsif (/^==========/ || /^changelog stats/) {
+                                $started = 0;
+                                }
+                        elsif ($started) {
+                                $cl .= $_."\n";
+                                }
+                        }
+                close(YUMCL);
+		}
+	elsif (!$cl && $software::yum_command =~ /dnf/) {
+		# Run dnf updateinfo for this package and version
+		&open_execute_command(DNFUI,
+			"$software::yum_command updateinfo info ".
+		        quotemeta($pkg->{'name'}), 1, 1);
+		while(<DNFUI>) {
 			s/\r|\n//g;
-			if (/^\Q$pkg->{'name'}-$pkg->{'version'}\E/) {
+			if (/^\s*Description\s*:\s*(.*)/) {
 				$started = 1;
+				$cl .= $1."\n";
 				}
-			elsif (/^==========/ || /^changelog stats/) {
+			elsif ($started && /^\s*:\s*(.*)/) {
+				$cl .= $1."\n";
+				}
+			else {
 				$started = 0;
 				}
-			elsif ($started) {
-				$cl .= $_."\n";
-				}
 			}
-		close(YUMCL);
-
+		close(DNFUI);
+		}
+	if ($cl) {
 		# Save the cache
 		if (!-d $yum_changelog_cache_dir) {
 			&make_dir($yum_changelog_cache_dir, 0700);
@@ -596,7 +668,7 @@ unlink($available_cache_file.'1');
 }
 
 # list_for_mode(mode, nocache)
-# If not is 'updates' or 'security', return just updates. Othewise, return
+# If not is 'updates' or 'security', return just updates. Otherwise, return
 # all available packages.
 sub list_for_mode
 {
@@ -605,15 +677,71 @@ return $mode eq 'updates' || $mode eq 'security' ?
 	&list_possible_updates($nocache) : &list_available($nocache);
 }
 
-# check_reboot_required(after-flag)
+# check_reboot_required()
 # Returns 1 if the package system thinks a reboot is needed
 sub check_reboot_required
 {
 if ($gconfig{'os_type'} eq 'debian-linux') {
-        return -e "/var/run/reboot-required" ? 1 : 0;
-        }
+	if (-e "/var/run/reboot-required") {
+		return 1;
+		}
+	if (&has_command("needrestart")) {
+		my $out = &backquote_command("needrestart -b -k -n 2>&1 </dev/null");
+		if ($out =~ /NEEDRESTART-KSTA:\s*(?<status>\d)/m) {
+			if ($+{status} == 3) {
+				return 1;
+				}
+			}
+		}
+	}
+elsif ($gconfig{'os_type'} eq 'redhat-linux') {
+	my $needs_restarting_cmd = "needs-restarting";
+	my $needs_restarting = has_command($needs_restarting_cmd);
+	if ($needs_restarting) {
+		my $needs_restarting_rs =
+			&backquote_command("$needs_restarting -r 2>&1 </dev/null");
+		return 1
+		    if ("$?" != 0 && index($needs_restarting_rs, $needs_restarting_cmd) == -1);
+		}
+	}
 return 0;
 }
 
+# start_update_progress(&packages)
+# Record that a bunch of package updates are in progress by this process
+sub start_update_progress
+{
+my ($pkgs) = @_;
+if (!-d $update_progress_dir) {
+	&make_dir($update_progress_dir, 0700);
+	}
+my $f = "$update_progress_dir/$$";
+&write_file($f, { 'pid' => $$,
+		  'pkgs' => join(' ', @$pkgs) });
+}
+
+# end_update_progress()
+# Clear update progress marker file
+sub end_update_progress
+{
+my $f = "$update_progress_dir/$$";
+&unlink_file($f);
+}
+
+# get_update_progress()
+# Returns a list of hash refs, one per update in progress
+sub get_update_progress
+{
+my @rv;
+foreach my $f (glob("$update_progress_dir/*")) {
+	my %u;
+	&read_file($f, \%u) || next;
+	$u{'pid'} || next;
+	kill(0, $u{'pid'}) || next;
+	push(@rv, \%u);
+	}
+return @rv;
+}
+
 1;
 
